from tensorflow.keras import layers, Model
import tensorflow as tf
import numpy as np

from tensorflow.keras.utils import plot_model

def Upsampling(inputs, factor=2, **kwargs):
    """ Upsample an NxNxM array by a factor of 2 in the first 2 dimensions (= 2Nx2NxM).
    Call this function twice to upsample from 2x to 4x, and call 3 times to upsample by x8 and so on.
    This function only upsamples spatially. The last dimension (filters or channels) remains the same.

    Parameters
    ----------
    inputs: a tf.keras.layer
    factor: scalar, minimum of 2
        The function multiplies the number of channels by factor^2.
    **kwargs: extra arguments to layers.Conv2D

    Returns
    -------
    x: a tf.nn layer.

    """
    x = layers.Conv2D(64 * (factor ** 2), 3, padding="same", **kwargs)(inputs)
    x = tf.nn.depth_to_space(x, block_size=factor)
    #x = layers.Conv2D(64 * (factor ** 2), 3, padding="same", **kwargs)(x)
    #x = tf.nn.depth_to_space(x, block_size=factor)
    return x

def ResBlock(inputs,Nfilters=64,dropoutRate=0,activations=['relu',None]):
    x = layers.Conv2D(Nfilters, 3, padding="same", activation=activations[0])(inputs)
    x = layers.Dropout(rate=dropoutRate)(x)
    x = layers.Conv2D(Nfilters, 3, padding="same",activation=activations[1])(x)
    x = layers.Dropout(rate=dropoutRate)(x)
    x = layers.Add()([inputs, x])
    return x   


def resnet_basic(Nfilters=64,kernelSize=3,NresBlocks=3,inputGroups=8,outputGroups=16,withIdentity=False,dropoutRate=0,activations=[None,None,None],resBlockActivations=['relu',None]):
    inputLayer = layers.Input(shape=(None,None,inputGroups))
    x = x_new = layers.Conv2D(Nfilters,kernelSize,padding="same",activation=activations[0])(inputLayer)
    # ResBlocks
    for _ in range(NresBlocks):
        x_new = ResBlock(x_new,Nfilters=Nfilters,dropoutRate=dropoutRate,activations=resBlockActivations)
    x_new = layers.Conv2D(Nfilters, kernelSize, padding="same",activation=activations[1])(x_new)
    x = layers.Add()([x, x_new])
    # upsampling block - each call to Upsampling will upsample by a factor of 2 in each direction.
    x = Upsampling(x)
    # increase channels (groups) to 16
    x = layers.Conv2D(outputGroups,3,padding="same",activation=activations[2])(x)
    # instantiate network
    if(withIdentity):
        identityLayerInput = layers.Input(shape=(None,None,outputGroups)) 
        identityLayerOutput = layers.Layer()(identityLayerInput) 
        x = layers.concatenate([x,identityLayerOutput])
        model = CustomModelUncertainty(inputs=[inputLayer,identityLayerInput],outputs=[x])
    else:
        model = Model(inputLayer,x)
    model.summary()
    #y_test = model.predict(moveaxis(X_val[0:1,:,:,:]))
    loss = 'mse' # works well
    #tf.keras.utils.plot_model(model,to_file='model.png',show_shapes=True)
    return model

class CustomModelUncertainty(tf.keras.Model):

    def train_step(self, data):
        """The logic for one training step.

        This method can be overridden to support custom training logic.
        For concrete examples of how to override this method see
        [Customizing what happens in fit](
        https://www.tensorflow.org/guide/keras/customizing_what_happens_in_fit).
        This method is called by `Model.make_train_function`.

        This method should contain the mathematical logic for one step of
        training.  This typically includes the forward pass, loss calculation,
        backpropagation, and metric updates.

        Configuration details for *how* this logic is run (e.g. `tf.function`
        and `tf.distribute.Strategy` settings), should be left to
        `Model.make_train_function`, which can also be overridden.

        Args:
          data: A nested structure of `Tensor`s.

        Returns:
          A `dict` containing values that will be passed to
          `tf.keras.callbacks.CallbackList.on_train_batch_end`. Typically, the
          values of the `Model`'s metrics are returned. Example:
          `{'loss': 0.2, 'accuracy': 0.7}`.
        """
        x, yerr, y = data 
        # Run forward pass.
        with tf.GradientTape() as tape:
            y_pred = self([x,yerr], training=True)
            loss = self.compiled_loss(y, y_pred)
        # Run backwards pass.
        self.optimizer.minimize(loss, self.trainable_variables, tape=tape)
        self.compiled_metrics.update_state(y, y_pred)
        # Collect metrics to return
        return_metrics = {}
        for metric in self.metrics:
            result = metric.result()
            if isinstance(result, dict):
                return_metrics.update(result)
            else:
                return_metrics[metric.name] = result
        return return_metrics

    def test_step(self, data):
        """The logic for one evaluation step.

        This method can be overridden to support custom evaluation logic.
        This method is called by `Model.make_test_function`.

        This function should contain the mathematical logic for one step of
        evaluation.
        This typically includes the forward pass, loss calculation, and metrics
        updates.

        Configuration details for *how* this logic is run (e.g. `tf.function` and
        `tf.distribute.Strategy` settings), should be left to
        `Model.make_test_function`, which can also be overridden.

        Args:
        data: A nested structure of `Tensor`s.

        Returns:
        A `dict` containing values that will be passed to
        `tf.keras.callbacks.CallbackList.on_train_batch_end`. Typically, the
        values of the `Model`'s metrics are returned.
        """
        
        x, yerr, y = data 

        y_pred = self([x,yerr], training=False)
        # Updates stateful loss metrics.
        self.compiled_loss(
            y, y_pred)
        self.compiled_metrics.update_state(y, y_pred)
        # Collect metrics to return
        return_metrics = {}
        for metric in self.metrics:
            result = metric.result()
            if isinstance(result, dict):
                return_metrics.update(result)
            else:
                return_metrics[metric.name] = result
        return return_metrics
