# %%
import numpy as np
import tensorflow as tf
import os, sys
import matplotlib.pyplot as plt
homeDir = os.path.expanduser("~")
# Edit this path to point to where your working copy of the flux-up code is stored.
sys.path.insert(0,homeDir + "/codes/flux-up")
sys.path.insert(0,homeDir + "/.local/lib/python3.9/site-packages")
import src
import h5py
import src.normalization as norm
sys.path.insert(0,'.')
from src.Pipeline import DynamicNormalizationPipeline as DynamicNormalizationPipeline
import pickle
import custom_model as cus
from scipy.io import savemat
from src.utilities import print_to_file
#%% Read in data
#runDirNumbers = range(6,7) # end dir not inclusive. 3, with only 100 tracks gives surprisingly good validation score
batchSize = 100
Nfilters = 256
trainingRuns = [7,8,10,13,14,46,15,16,17,18,47,28,29,30,31,48,34,39,40,49,50,41,42,43,44,45,35,36,37,38,51]
#trainingRuns = [6,11,27]
validationRuns = [6,11,9,12]
trainDatapath = ['../../OpenMC-Runs/{:04d}'.format(num) for num in trainingRuns]
valDatapath = ['../../OpenMC-Runs/{:04d}'.format(num) for num in validationRuns]
with open('openmc-runs.pickle', 'wb') as handle:
    pickle.dump({'trainingRuns':trainingRuns,'validationRuns':validationRuns}, handle, protocol=pickle.HIGHEST_PROTOCOL)

#%% set up pipeline
pl =  src.SplitRelativeErrorBasicPipeline()
train_ds, output_signature, val_ds, normX, normY = pl.connect(trainDatapath,valDatapath)
train_ds = train_ds.map(lambda X,Y,Xerr,Yerr: (tf.squeeze(X),tf.squeeze(Y),tf.squeeze(Xerr),tf.squeeze(Yerr)))
train_ds = train_ds.batch(batchSize)
#%% attempt to concatenate inputs and outputs correctly
def concatDs(X,Y,XrelErr,YrelErr):
    nan_mask = tf.math.is_nan(XrelErr)
    XrelErr = tf.where(nan_mask, tf.ones_like(XrelErr) * 1.0, XrelErr)
    nan_mask = tf.math.is_nan(YrelErr)
    YrelErr = tf.where(nan_mask, tf.ones_like(YrelErr) * 1.0, YrelErr)
    return X,YrelErr,tf.concat([Y,YrelErr],axis=-1)

train_ds_concat = train_ds.map(lambda X,Y,XrelErr,YrelErr: concatDs(X,Y,XrelErr,YrelErr))
val_ds_concat = val_ds.map(lambda X,Y,XrelErr,YrelErr: concatDs(X,Y,XrelErr,YrelErr))


#%% define loss function: Gaussian NLL
def customLoss(true,pred):
    true, _ = tf.split(true,2,axis=-1)
    pred, relErr = tf.split(pred,2,axis=-1)
    nan_mask = tf.math.is_nan(relErr)
    relErr = tf.where(nan_mask, tf.ones_like(relErr) * 1.0, relErr)
    sigma = tf.multiply(relErr,true) # supplied errors are relative, turn to absolute
    sigma = tf.where(sigma < 1e-7, tf.ones_like(relErr) * 1e-7, sigma)
    logprob = -tf.math.log(sigma) - 0.5*tf.math.log(2*np.pi) - 0.5*((true-pred)/sigma)**2
    logprob = -tf.math.abs(true-pred)/sigma
    loss = tf.reduce_mean(-logprob)
    return loss



def loss_mae_identity(true,pred):
    true, sigma = tf.split(true,2,axis=-1)
    pred, sigma = tf.split(pred,2,axis=-1)
    #print(true)
    loss = tf.reduce_mean(tf.math.abs(true - pred))
    return loss
# % instantiate model
#from src.ml_models import resnet_basic
activations = 3*['elu'] #['relu','relu','relu']
resBlockActivations = 2*['elu'] #['relu','relu']

model = cus.resnet_basic(Nfilters=Nfilters,kernelSize=3,NresBlocks=3,inputGroups=8,
    outputGroups=16,withIdentity=True,dropoutRate=0.0,activations=activations,
    resBlockActivations=resBlockActivations)

optim_edsr = tf.keras.optimizers.Adam(
    learning_rate=tf.keras.optimizers.schedules.PiecewiseConstantDecay(
        #boundaries=[1900,1900], values=[1e-4, 1e-5, 2e-6]
        boundaries=[int(200000/batchSize),int(200000/batchSize),int(400000/batchSize), int(800000/batchSize)], 
        values=[1e-4, 1e-4, 2e-5, 1e-5, 2e-6]
    )
)
metrics=[tf.keras.metrics.MeanAbsolutePercentageError(), 'mae']

esCallback = tf.keras.callbacks.EarlyStopping(patience = 100, monitor = 'val_loss')
cpCallback = tf.keras.callbacks.ModelCheckpoint('checkpoint-model-file', 
                    monitor="val_loss", mode="min", 
                    save_best_only=True, verbose=1)
model.compile(optimizer=optim_edsr, loss=loss_mae_identity, metrics=metrics)
history = model.fit(train_ds_concat,epochs=2500, validation_data=val_ds_concat,callbacks = [cpCallback])
# %% save model
model.save('./ml-model')

# %% export training results/history to txt and matlab files
endLoss = history.history['loss'][-1]
endValLoss = min(history.history['val_loss'])

print_to_file('final losses: ' + '{:#.3g}'.format(endLoss) + '/' + '{:#.3g}'.format(endValLoss),'loss.txt')

savemat('norm.mat', {'normX':normX.numpy(),'normY':normY.numpy()})
savemat('history.mat',{'train_mae':history.history['mae'],'train_mape':history.history['mean_absolute_percentage_error'],
    'val_mae':history.history['val_mae'],'val_mape':history.history['val_mean_absolute_percentage_error']})


# %%
